#include "types.h"
#include "sort.h"
#include "stdlib.h"
#include "error.h"

//  
#define T 6

//  
int* tape[T];
//  \
int* tape_ptr[T];

//   

void tapewrite(int tapeindex, int what)
{
  *tape_ptr[tapeindex]=what;
  tape_ptr[tapeindex]++;
}

//   

int taperead(int tapeindex)
{
  return *tape_ptr[tapeindex]++;
}

//  

void tapeinit(int tapeindex)
{
  tape_ptr[tapeindex]=tape[tapeindex];
}

//  

void tapesinit()
{
  int i=T;
  for(i=0;i<T;i++)
    tapeinit(i);
}

//    ,    

void tapescreate(int datasize)
{
  int i;
  for(i=0;i<T;i++)
  {
    tape[i]=(int*)malloc(datasize*sizeof(int));
    tape_ptr[i]=tape[i];
  }
}

//  

void tapesclose()
{
  int i;
  for(i=0;i<T;i++)
    free(tape[i]);
}

//   
//      (-1 0 1)
int compare(field* data, int idx1,int idx2)
{
  const char* p1,*p2;

  if(idx1==-1&&idx2==-1)
    return 0;
  if(idx1==-1)
    return 1;
  if(idx2==-1)
    return -1;
  if(idx1==idx2)
    return 0;

  p1=data[idx1];
  p2=data[idx2];
  return _stringformatcmp(p1,p2);
  /*
  if(*p1=='"'&&*p2=='"')
  {//  
    return _strcmp(p1,p2);
  }else if(*p1!='"'&&*p1!='"')
  {//  
    double fp1=atof(p1),fp2=atof(p2);
    if(fp1==fp2)
      return 0;
    return fp1<fp2?-1:1;
  }else
  {//   ,     
    return 0;
  }
  */
}

//  

void sortfields(field* data,int* idx,int datacount)
{
  int i,ii;
restart:
  for(i=0;i<datacount;i++)
  {
    for(ii=i+1;ii<datacount;ii++)
    {
      if(idx[i]==-1||idx[ii]==-1)
        continue;

      if(compare(data,idx[i],idx[ii])>0)
      {
        int bak;

        bak=idx[i];
        idx[i]=idx[ii];
        idx[ii]=bak;

        goto restart;
      }
    }
  }
}


//     

int taperead_f(int *D,int *F,int *A,int *C,int *tape_logical,int tapeindex)
{
  int value;
  C[tape_logical[tapeindex]]++;
  D[tape_logical[tapeindex]]--;

  value=taperead(tape_logical[tapeindex]);
  if(value==-1)
    F[tape_logical[tapeindex]]--;

  return value;
}

//   ,    

void tapewrite_f(int*D,int *F,int *tape_logical,int tapeindex,int value)
{
  D[tape_logical[tapeindex]]++;
  if(value==-1)
  {
    F[tape_logical[tapeindex]]++;
  }

  {
    tapewrite(tape_logical[tapeindex],value);
  }
}

// D[] -     ( )
// F[] -   
// A[] -    
// C[] -   (  )
// n -  ,        

//    .       
//    1,      

int cascadedistribution(field* data,int *D,int *F,int *A,int *C,int n,int *tape_logical,int tapecount)
{
  int tapeindex;
  int min;
  int *tapebuf=malloc(sizeof(int)*tapecount);
  int *tapeaccess=malloc(sizeof(int)*tapecount);
  int ret = 0;

  for(tapeindex=0;tapeindex<n;tapeindex++)
  {
    tapebuf[tape_logical[tapeindex]]=taperead_f(D,F,A,C,tape_logical,tapeindex);
    tapeaccess[tape_logical[tapeindex]]=1;

    if(D[tape_logical[tapeindex]]==0)
      ret++;
  }

  {
loop:
    min=-1;

    // min          .
    //         ,        
    //   C[] < A[] (     )

    //           
    //    

    for(tapeindex=0;tapeindex<n;tapeindex++)
    {
      if(!tapeaccess[tape_logical[tapeindex]])
        continue;

      if(min==-1)
      {
        min=tapeindex;
        continue;
      }

      if(compare(data,tapebuf[tape_logical[tapeindex]],tapebuf[tape_logical[min]])<0)
        min=tapeindex;
    }

    //   
    if(min!=-1)
    {
      //       
      tapewrite_f(D,F,tape_logical,n,tapebuf[tape_logical[min]]);
      //          
      //     .
      //        ,   
      //         (  
      //      ,     )
      if(C[tape_logical[min]]<A[tape_logical[min]])
      {
        tapebuf[tape_logical[min]]=taperead_f(D,F,A,C,tape_logical,min);
        if(D[tape_logical[min]]==0)
          ret++;
      }else
        tapeaccess[tape_logical[min]]=0;
      goto loop;
    }
  }

  free(tapebuf);
  free(tapeaccess);

  return ret;
}

// n- 
//  

//       
//      1

int cascadepathlink(field* data,int*D,int *F,int *A,int n,int *tape_logical,int tapecount)
{
  // n =  .
  // n-1 -  

  int * C=malloc(sizeof(int)*tapecount);
  //int i;
  int ret=0;
  int code=0;

  while(1)
  {
    _memset(C,0,sizeof(int)*tapecount);
    switch(cascadedistribution(data,D,F,A,C,n,tape_logical,tapecount))
    {
    case 0:
      break; // goto looop
    case 1:
      goto exit;
    default: //if(code>1)
      ret=1;
      goto exit;
    }
  }

exit:
  free(C);
  return ret;
}

//   ,     

void cascadepass(field* data,int *D,int *F,int* A,int *tape_logical,int tapecount)
{
  int i;

  for(i=tapecount-1;i>0;i--)
  {
    tapeinit(tape_logical[i]);
    if(cascadepathlink(data,D,F,A,i,tape_logical,tapecount))
      break;
  }
}

#ifdef _DEBUG

void sprintf(char *, const char *, ...);

void tapesdump(field* data,int *D,int *F,int* A,int *tape_logical,int tapecount)
{
  int casette;

  tapesinit();

  for(casette=0;casette<T;casette++)
  {
    int element;
    char buf[1024];
    sprintf(buf,"Casette %d (: %d, : %d, : %d)\n",casette,D[casette],F[casette],A[casette]);
    TRACE(buf);
    for(element=0;element<D[casette];element++)
    {
      int value=taperead(casette);
      sprintf(buf,"\t%d - '%s'(idx:%d)\n",element,value==-1?"":data[value],value);
      TRACE(buf);
      if(element+1%A[casette]==0)
        TRACE("\n");
    }
  }
}

#else

#define tapesdump(data,D,F,A,tape_logical,tapecount) (void)0

#endif // _DEBUG

//   

int fibonacheinit(int *f,int *tape_logical,int tapecount)
{
  int i,total=0;

  for(i=0;i<tapecount-1;i++)
  {
    f[tape_logical[i]]=1;
    total+=f[tape_logical[i]];
  }
  return total;
}

//   

int fibonachenext(int *f,int *tape_logical,int tapecount)
{
  //   0  1  2  3  4  5

  // physical

  //   0  0  0  0  0  0
  //   1  1  1  1  1  -
  //   -  1  2  3  4  5
  //   15 14 12 9  5  -

  // logical

  //   0  0  0  0  0  0
  //   1  1  1  1  1  -
  //   5  4  3  2  1  -
  //   15 14 12 9  5  -

  int i,total=0;
  int* fnew;

  fnew=malloc(sizeof(int)*tapecount);
  _memset(fnew,0,sizeof(int)*tapecount);

  //      f[tape_logical[0]] 
  //      
  fnew[tape_logical[1]]=f[tape_logical[0]];
  total=fnew[tape_logical[1]];

  for(i=2;i<tapecount;i++)
  {
    fnew[tape_logical[i]]=fnew[tape_logical[i-1]]+f[tape_logical[i-1]];
    total+=fnew[tape_logical[i]];
  }

  _memcpy(f,fnew,sizeof(int)*tapecount);

  free(fnew);

  return total;
}

//   

void tapesinvert(int *tape_logical,int tapecount)
{
  int i;

  for(i=0;i<tapecount/2;i++)
  {
    int bak;

    bak=tape_logical[i];
    tape_logical[i]=tape_logical[tapecount-1-i];
    tape_logical[tapecount-1-i]=bak;
  }
}

//   

void tapesreset(int *tape_logical,int tape_count)
{
  int i;
  for(i=0;i<tape_count;i++)
    tape_logical[i]=i;
}

int sort(field *data,int datasize,int *selectfieldsidx,int selectfieldscount,int orderfieldidx)
{
  const int linecount = datasize/selectfieldscount;
  //  
  int tape_logical[T];
  //   
  int l = 0;
  //      
  int D[T]={0};
  //   
  int F[T]={0};
  int keyfieldnum=findinnumarray(selectfieldsidx,selectfieldscount,orderfieldidx);

  tapescreate(1000);
  tapesreset(tape_logical,T);

  //  

  {
    int A[T]={0};
    int offset=0;

    //    ,   
    int total=fibonacheinit(A,tape_logical,T);
    l=1;
    while(total<linecount)
    {
      total=fibonachenext(A,tape_logical,T);
      tapesinvert(tape_logical,T);
      l++;
    }
    for(offset=keyfieldnum;offset<datasize;offset+=selectfieldscount)
    {
      //      
      {
        int i;
        for(i=0;i<T;i++)
        {
          if(D[tape_logical[i]]<A[tape_logical[i]])
          {
            tapewrite_f(D,F,tape_logical,i,offset);
            break;
          }
        }
      }
    }

    //       
    //     ,    D (
    //       ) 
    //    

    {
      int i,offset;
      for(offset=linecount;offset<total;offset++)
      {
        for(i=0;i<T;i++)
        {
          if(D[tape_logical[i]]<A[tape_logical[i]])
          {
            tapewrite_f(D,F,tape_logical,i,-1);
          }
        }
      }
    }
  }

  //  

  {
    //     ,    
    //   .
    int A[T]={0};
    fibonacheinit(A,tape_logical,T);

    while(l--)
    {
      tapesdump(data,D,F,A,tape_logical,T);

      tapesinit();
      cascadepass(data,D,F,A,tape_logical,T);
      fibonachenext(A,tape_logical,T);
      tapesinvert(tape_logical,T);

      tapesdump(data,D,F,A,tape_logical,T);
    }
  }

  //  

  {
    int index=0;
    field *datanew=malloc(sizeof(field)*datasize);

    tapesinit();
    while(index<linecount)
    {
      int i;
      int idx=taperead(tape_logical[0]);

      for(i=0;i<selectfieldscount;i++)
        _strcpy(datanew[index*selectfieldscount+i],data[idx-keyfieldnum+i]);
      index++;
    }
    _memcpy(data,datanew,sizeof(field)*datasize);
    free(datanew);
  }

  tapesclose();
  return 0;
}
